package dev.fitko.fitconnect.core.http;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.github.tomakehurst.wiremock.WireMockServer;
import com.github.tomakehurst.wiremock.client.WireMock;
import dev.fitko.fitconnect.api.config.build.BuildInfo;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.exceptions.internal.RestApiException;
import okhttp3.RequestBody;
import okhttp3.Response;
import okhttp3.ResponseBody;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.net.URI;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import static com.github.tomakehurst.wiremock.client.WireMock.badRequest;
import static com.github.tomakehurst.wiremock.client.WireMock.equalTo;
import static com.github.tomakehurst.wiremock.client.WireMock.getRequestedFor;
import static com.github.tomakehurst.wiremock.client.WireMock.matching;
import static com.github.tomakehurst.wiremock.client.WireMock.ok;
import static com.github.tomakehurst.wiremock.client.WireMock.urlEqualTo;
import static com.github.tomakehurst.wiremock.client.WireMock.urlPathEqualTo;
import static com.github.tomakehurst.wiremock.client.WireMock.verify;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.hamcrest.Matchers.notNullValue;
import static org.hamcrest.Matchers.nullValue;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class HttpClientTest {

    private final WireMockServer wireMockServer = new WireMockServer();

    @BeforeEach
    public void startWireMock() {
        this.wireMockServer.start();
    }

    @AfterEach
    public void shutDownWireMock() {
        this.wireMockServer.stop();
    }

    @Test
    public void createRequestBodyWithImplicitContentType() throws JsonProcessingException {

        // When
        RequestBody requestBody = new HttpClient().createRequestBody(Map.of(), "abc");

        // Then
        assertThat(Objects.requireNonNull(requestBody.contentType()).toString(), is(MimeTypes.APPLICATION_JSON));
    }

    /**
     * See comments at {@link HttpClient#createRequestBody(Map, Object)} for more information
     */
    @Test
    public void createRequestBodyWithNoQuotesInString() throws IOException {

        // When
        RequestBody requestBody = new HttpClient().createRequestBody(Map.of(), "abc");

        // Then
        assertThat(requestBody.contentLength(), is(3L));
    }

    @Test
    public void createRequestBodyWithExplicitContentType() throws IOException {

        // When
        RequestBody requestBody = new HttpClient().createRequestBody(
                Map.of("Content-Type", "application/json"), 10);

        // Then
        assertThat(requestBody.contentLength(), is(2L));
        assertThat(Objects.requireNonNull(requestBody.contentType()).toString(), is("application/json"));
    }

    @Test
    public void respondWithRestApiExceptionIfConfigured() {

        // Given
        Response response = mock(Response.class);
        when(response.code()).thenReturn(400);
        when(response.isSuccessful()).thenReturn(false);

        // When
        RestApiException exception = Assertions.assertThrows(RestApiException.class,
                () -> new HttpClient().evaluateStatusAndRespond(response, String.class));

        // Then
        assertThat(exception.getStatusCode(), is(400));
    }

    @Test
    public void respondWithNoExceptionIfNotConfigured() throws IOException {

        // Given
        Response response = mock(Response.class);
        when(response.code()).thenReturn(400);
        when(response.isSuccessful()).thenReturn(false);
        when(response.body()).thenReturn(mock(ResponseBody.class));

        // When
        HttpResponse<?> httpResponse = new HttpClient(false).evaluateStatusAndRespond(response, String.class);

        // Then
        assertThat(httpResponse.getStatusCode(), is(400));
    }

    @Test
    public void respondWithIOExceptionIfResponseBodyIsNull() {

        // Given
        Response response = mock(Response.class);
        when(response.isSuccessful()).thenReturn(true);

        // When
        RestApiException exception = Assertions.assertThrows(RestApiException.class,
                () -> new HttpClient().evaluateStatusAndRespond(response, String.class));

        // Then
        assertThat(exception.getMessage(), is("Response body is null."));
    }

    @Test
    public void respondWithEmptyBodyIfResponseTypeIsVoid() throws IOException {

        // Given
        Response response = mock(Response.class);
        when(response.code()).thenReturn(200);
        when(response.isSuccessful()).thenReturn(true);
        when(response.body()).thenReturn(mock(ResponseBody.class));

        // When
        HttpResponse<?> httpResponse = new HttpClient().evaluateStatusAndRespond(response, Void.class);

        // Then
        assertThat(httpResponse.getBody(), nullValue());
        assertThat(httpResponse.getStatusCode(), is(200));
    }

    @Test
    public void respondWithRawStringIfResponseTypeIsVoid() throws IOException {

        // Given
        ResponseBody responseBody = mock(ResponseBody.class);
        when(responseBody.string()).thenReturn("abc");
        Response response = mock(Response.class);
        when(response.code()).thenReturn(200);
        when(response.isSuccessful()).thenReturn(true);
        when(response.body()).thenReturn(responseBody);

        // When
        HttpResponse<?> httpResponse = new HttpClient().evaluateStatusAndRespond(response, String.class);

        // Then
        assertThat(httpResponse.getBody(), is("abc"));
        assertThat(httpResponse.getStatusCode(), is(200));
    }

    @Test
    public void respondWithComplexBodyIfResponseTypeComplex() throws IOException {

        // Given
        ResponseBody responseBody = mock(ResponseBody.class);
        when(responseBody.string()).thenReturn("{ \"property\": \"value\" }");
        Response response = mock(Response.class);
        when(response.code()).thenReturn(200);
        when(response.isSuccessful()).thenReturn(true);
        when(response.body()).thenReturn(responseBody);

        // When
        HttpResponse<?> httpResponse = new HttpClient().evaluateStatusAndRespond(response, Map.class);

        // Then
        assertThat(httpResponse.getBody(), is(Map.of("property", "value")));
        assertThat(httpResponse.getStatusCode(), is(200));
    }

    @Test
    public void get() {

        // Given
        WireMock.configureFor("localhost", 8080);
        this.wireMockServer.stubFor(WireMock.get(urlEqualTo("/test"))
                .withHeader("User-Agent", equalTo("online service"))
                .willReturn(ok().withBody("success"))
        );

        // When
        HttpResponse<String> httpResponse = new HttpClient()
                .get("http://localhost:8080/test", Map.of("User-Agent", "online service"), String.class);

        // Then
        assertThat(httpResponse.getStatusCode(), is(200));
        assertThat(httpResponse.getBody(), is("success"));
    }

    @Test
    public void post() {

        // Given
        WireMock.configureFor("localhost", 8080);
        this.wireMockServer.stubFor(WireMock.post(urlEqualTo("/test"))
                .withHeader("User-Agent", equalTo("online service"))
                .withRequestBody(matching("data"))
                .willReturn(ok().withBody("success"))
        );

        // When
        HttpResponse<String> httpResponse = new HttpClient()
                .post("http://localhost:8080/test", Map.of("User-Agent", "online service"), "data", String.class);

        // Then
        assertThat(httpResponse.getStatusCode(), is(200));
        assertThat(httpResponse.getBody(), is("success"));
    }

    @Test
    public void put() {

        // Given
        WireMock.configureFor("localhost", 8080);
        this.wireMockServer.stubFor(WireMock.put(urlEqualTo("/test"))
                .withHeader("User-Agent", equalTo("online service"))
                .withRequestBody(matching("data"))
                .willReturn(ok().withBody("success"))
        );

        // When
        HttpResponse<String> httpResponse = new HttpClient()
                .put("http://localhost:8080/test", Map.of("User-Agent", "online service"), "data", String.class);

        // Then
        assertThat(httpResponse.getStatusCode(), is(200));
        assertThat(httpResponse.getBody(), is("success"));
    }

    @Test
    public void ignoreUnknownProperties() throws JsonProcessingException {

        // Given
        WireMock.configureFor("localhost", 8080);
        this.wireMockServer.stubFor(WireMock.get(urlEqualTo("/test"))
                .withHeader("User-Agent", equalTo("online service"))
                .willReturn(ok().withBody(new ObjectMapper().writeValueAsString(getMetadataWithUnknownSchemaProperty())))
        );

        // When
        HttpResponse<Metadata> httpResponse = new HttpClient()
                .get("http://localhost:8080/test", Map.of("User-Agent", "online service"), Metadata.class);

        // Then
        assertThat(httpResponse.getStatusCode(), is(200));
        assertThat(httpResponse.getBody(), notNullValue());
    }

    @Test
    public void userAgentInterceptorsIsSetCorrect() {

        // Given
        WireMock.configureFor("localhost", 8080);
        this.wireMockServer.stubFor(WireMock.get(urlEqualTo("/test"))
                .withHeader("User-Agent", equalTo("productName/productVersion (commit:commit;os:" + System.getProperty("os.name") + ")"))
                .willReturn(ok())
        );

        BuildInfo buildInfo = new BuildInfo();
        buildInfo.setCommit("commit");
        buildInfo.setProductName("productName");
        buildInfo.setProductVersion("productVersion");

        // When
        HttpResponse<String> httpResponse = new HttpClient(List.of(new UserAgentInterceptor(buildInfo)))
                .get("http://localhost:8080/test", String.class);

        // Then
        verify(getRequestedFor(urlPathEqualTo("/test"))
                .withHeader("User-Agent", equalTo("productName/productVersion (commit:commit;os:" + System.getProperty("os.name") + ")")));
        assertThat(httpResponse.getStatusCode(), is(200));
    }

    @Test
    public void exceptionHasResponseBodyAsMessage() {

        // Given
        WireMock.configureFor("localhost", 8080);
        this.wireMockServer.stubFor(WireMock.get(urlEqualTo("/test")).willReturn(badRequest().withBody("failure")));

        // When
        RestApiException exception = Assertions.assertThrows(RestApiException.class,
                () -> new HttpClient().get("http://localhost:8080/test", String.class));

        // Then
        assertThat(exception.getStatusCode(), is(400));
        assertThat(exception.getMessage(), is("failure"));
    }

    private static Map<String, Map<String, Object>> getMetadataWithUnknownSchemaProperty() {
        return Map.of(
                "$schemaFoo", Collections.emptyMap(),
                "contentStructure", Map.of(
                        "data", Map.of(
                                "hash", Map.of(
                                        "type", SignatureType.SHA_512,
                                        "content", "bla"
                                ),
                                "submissionSchema", Map.of(
                                        "schemaUri", URI.create("https://dummy.schema.url"),
                                        "mimeType", MimeType.APPLICATION_JSON
                                )
                        )
                )
        );
    }
}
